import abc
from abc import ABC
from pathlib import Path
import re
from typing import List, NamedTuple, Final, Union, Iterable, Optional, Literal, Tuple

NUMBERED_ITEM_MAX_DIGITS = 6
SPECIAL_FRAME_SEQUENCE_ITEMS: Final[List[str]] = ['data', 'data_syn']

# template still contain substring which have to be replaced before usage as a pattern like 'n'
NUMBERED_ITEM_TEMPLATE = r'[0-9]{1,n}'
# using fr, backslashes do not have to be escaped in f-strings
LIST_TEMPLATE = fr"\[{NUMBERED_ITEM_TEMPLATE}(,{NUMBERED_ITEM_TEMPLATE})*\]"
# 'item' can be replaced by 'sequence' or 'frame', which is necessary when the
# template is used twice in a pattern, since capture groups have to have unique names
RANGE_TEMPLATE = f"(?P<itemstart>{NUMBERED_ITEM_TEMPLATE})?:(?P<itemstop>{NUMBERED_ITEM_TEMPLATE})?" + \
                 f"(:(?P<itemstep>-?{NUMBERED_ITEM_TEMPLATE})?)?"
STAR_TEMPLATE = r'\*'

SINGLE_ITEM_PATTERN: Final[re.Pattern] = re.compile(r'^(?P<item>[0-9]+|\*|data|data_syn)$')
LIST_PATTERN: Final[re.Pattern] = re.compile(
    fr'^\[(?P<items>{NUMBERED_ITEM_TEMPLATE}(,{NUMBERED_ITEM_TEMPLATE})*)\]$'.replace(
        'n', str(NUMBERED_ITEM_MAX_DIGITS)))
RANGE_PATTERN: Final[re.Pattern] = re.compile(RANGE_TEMPLATE.replace('n', str(NUMBERED_ITEM_MAX_DIGITS)))


class RangeSelection(ABC):
    def __init__(self, start: Optional[str], stop: Optional[str], step: Optional[Union[int, str]]):
        self.start = self.check_start(start) if start else None
        self.stop = self.check_stop(stop) if stop else None
        self.step = RangeSelection.check_step(step) if step else None

    @abc.abstractmethod
    def check_start(self, start: Optional[str]) -> str:
        pass

    @abc.abstractmethod
    def check_stop(self, stop: Optional[str]) -> str:
        pass

    def as_tuple(self) -> Tuple[Optional[str], Optional[str], Optional[int]]:
        return self.start, self.stop, self.step

    @staticmethod
    def check_step(step: Optional[Union[int, str]]) -> Optional[int]:
        if not isinstance(step, (int, str)):
            raise TypeError("step has to be of type 'int' or 'str'")

        if isinstance(step, str):
            # consider a possible minus sign
            if not step.lstrip('-').isdigit():
                raise ValueError(f"step represents no digit: {step}")
            step = int(step)

        if step == 0:
            raise ValueError('step must not be 0')

        return step


class SequenceRangeSelection(RangeSelection):
    def __init__(self, start: Optional[str], stop: Optional[str], step: Optional[Union[int, str]]):
        super().__init__(start, stop, step)

    def check_start(self, start):
        return normalize_numbered_sequence_selection_item(start)

    def check_stop(self, stop):
        return normalize_numbered_sequence_selection_item(stop)


class FrameRangeSelection(RangeSelection):
    def __init__(self, start: Optional[str], stop: Optional[str], step: Optional[Union[int, str]]):
        super().__init__(start, stop, step)

    def check_start(self, start):
        return normalize_numbered_frame_selection_item(start)

    def check_stop(self, stop):
        return normalize_numbered_frame_selection_item(stop)


def get_selection_expression_pattern() -> re.Pattern:
    # # most general pattern, valid for a frame sequence selection or a frame selection
    selection_template = f"{NUMBERED_ITEM_TEMPLATE}|{LIST_TEMPLATE}|{RANGE_TEMPLATE}|{STAR_TEMPLATE}"

    frame_selection_pattern = selection_template.replace('n', '6').replace('item', 'frame')

    frame_sequence_selection_pattern = selection_template.replace('n', '4').replace('item', 'sequence')
    for item in SPECIAL_FRAME_SEQUENCE_ITEMS:
        frame_sequence_selection_pattern += f"|{item}"

    return re.compile(
        f"^(?P<framesequence>{frame_sequence_selection_pattern})/(?P<frame>{frame_selection_pattern})$"
    )


SELECTION_EXPRESSION_PATTERN: Final[re.Pattern] = get_selection_expression_pattern()


class FrameSelector(NamedTuple):
    frame_sequence_selection: str
    frame_selection: str


def get_frame_selectors(selections: List[str]) -> List[FrameSelector]:
    frame_selectors = []

    for index, selection in enumerate(selections):
        try:
            frame_selectors.append(get_frame_selector(selection))
        except ValueError as error:
            raise ValueError(
                f"Selection syntax is incorrect: Selection '{selection}' at index {index}") from error

    return frame_selectors


def is_star_selection(selection: List[str]) -> bool:
    return len(selection) == 1 and selection[0] == '*'


def get_frame_selector(selection: str) -> FrameSelector:
    if match := SELECTION_EXPRESSION_PATTERN.match(selection):
        frame_sequence_selection = match.group('framesequence')
        frame_selection = match.group('frame')
# TODO validate range expression
        return FrameSelector(frame_sequence_selection, frame_selection)
    else:
        raise ValueError(f"Selection syntax is incorrect: '{selection}'")


def get_items(selection: str) -> List[str]:
    items = []
    if match := LIST_PATTERN.match(selection):
        items = [item for item in match.group('items').split(',')]
    elif match := SINGLE_ITEM_PATTERN.match(selection):
        items.append(match.group('item'))
    else:
        raise ValueError(f"Descriptor syntax is incorrect: {selection}")

    return items


def range_selection(selection: str, kind: Literal['sequence', 'frame']) -> Optional[RangeSelection]:
    if match := RANGE_PATTERN.match(selection):
        start = match.group('itemstart')
        stop = match.group('itemstop')
        step = match.group('itemstep')

        if kind == 'sequence':
            return SequenceRangeSelection(start, stop, step)
        elif kind == 'frame':
            return FrameRangeSelection(start, stop, step)

    return None


def normalize_sequence_selection_item(item: str) -> str:
    if not isinstance(item, str):
        raise TypeError('item must be a string')

    if item in ['*', *SPECIAL_FRAME_SEQUENCE_ITEMS]:
        return item

    try:
        return normalize_numbered_sequence_selection_item(item)
    except ValueError:
        raise ValueError(f"Invalid frame sequence selection item: {item}")


def normalize_numbered_sequence_selection_item(item: str) -> str:
    if item.isdigit() and len(item) <= 4:
        return "{:04d}".format(int(item))

    raise ValueError(
        f"Not a numbered frame sequence selection item: {item} (Must represent a digit and of length <= 4)")


def normalize_frame_selection_item(item: str) -> str:
    if not isinstance(item, str):
        raise TypeError('item must be a string')

    if item == '*':
        return item

    try:
        return normalize_numbered_frame_selection_item(item)
    except ValueError:
        raise ValueError(f"Invalid frame selection item: {item}")


def normalize_numbered_frame_selection_item(item: str) -> str:
    if item.isdigit() and len(item) <= 6:
        return "{:06d}".format(int(item))

    raise ValueError(
        f"Not a numbered frame selection item: {item} (Must represent a digit and of length <= 6)")


def load_frame_selectors_from_file(file: Union[Path, str]) -> Iterable[FrameSelector]:
    file_path = Path(file)

    if not file_path.exists():
        raise IOError(f"File does not exist: {file_path}")

    selectors = []
    with open(file_path, 'r') as f:
        for line_number, line in enumerate(f):
            # skip empty lines
            if line.rstrip():
                try:
                    selectors.append(get_frame_selector(line))
                except ValueError as error:
                    raise ValueError(f"Invalid selection syntax at line {line_number} in {file_path}") from error

    return selectors



import functools
import os
from pathlib import Path
import random
from typing import List, Iterable, Union, Iterator, Dict, Optional

from . import datatypes
from . import frameselection
from . import utils


class YcbVideoLoader:
    """

    """
    def __init__(self, path: Union[Path, str]):
        self._path = utils.validate_directory_path(path)
        self._data_directory = self._path / 'data'
        self._data_syn_directory = self._path / 'data_syn'
        self._available_frame_sequences = self.get_available_frame_sequences()
    """
        path: Path to the YCB-Video dataset root directory
    """

    def frames_info(self) -> Dict[str, Dict[str, Optional[List[str]]]]:
        available_sequences = sorted(self.get_available_frame_sequences())

        info = {}
        for sequence_identifier in available_sequences:
            sequence = self.get_frame_sequence(sequence_identifier)
            complete_frame_sets, incomplete_frame_sets = sequence.get_available_frame_sets()

            frame_sets = {}
            for complete_set in complete_frame_sets:
                frame_sets[complete_set] = None
            for identifier, incomplete_set in incomplete_frame_sets.items():
                frame_sets[identifier] = incomplete_set

            info[sequence_identifier] = frame_sets

        return info

    def get_available_data_frame_sequences(self) -> List[str]:
        return os.listdir(self._data_directory)

    def get_available_frame_sequences(self) -> List[str]:
        available_sequences = self.get_available_data_frame_sequences()

        if self._data_syn_directory.exists():
            available_sequences.append(self._data_syn_directory.name)

        return available_sequences

    @functools.lru_cache()
    def get_frame_sequence(self, index: Union[str, int]) -> datatypes.FrameSequence:
        if isinstance(index, int):
            index = "{:04d}".format(index)

        path = self._data_syn_directory if index == self._data_syn_directory.name else self._data_directory / index

        if index not in self.get_available_frame_sequences():
            raise IOError(f"Frame sequence does not exist: {index} ({path})")

        return datatypes.FrameSequence(path)

    def get_frame_sequences(self, indexes: Iterable[Union[int, str]]) -> List[datatypes.FrameSequence]:
        return [self.get_frame_sequence(index) for index in indexes]

    def get_frame_descriptors(self, frame_selector: frameselection.FrameSelector) -> List[datatypes.FrameDescriptor]:
        frame_descriptors = []

        actual_frame_sequence_selection = None
        actual_frame_selection = None

        if frame_selector.frame_sequence_selection == '*':
            actual_frame_sequence_selection = sorted(self._available_frame_sequences)
        elif frame_selector.frame_sequence_selection == 'data':
            actual_frame_sequence_selection = sorted((self.get_available_data_frame_sequences()))
        elif sequence_range := frameselection.range_selection(frame_selector.frame_sequence_selection, 'sequence'):
            try:
                actual_frame_sequence_selection = utils.expand_range(
                    sequence_range.as_tuple(),
                    sorted(self.get_available_data_frame_sequences()))
            except utils.MissingItemError as error:
                raise IOError(
                    f"Frame sequence specified in range as '{error.usage}' is not available: '{error.item}'") from error
        else:
            actual_frame_sequence_selection = \
                [frameselection.normalize_sequence_selection_item(frame_sequence) for
                 frame_sequence in frameselection.get_items(frame_selector.frame_sequence_selection)]
            for index, frame_sequence in enumerate(actual_frame_sequence_selection):
                if frame_sequence not in self._available_frame_sequences:
                    raise IOError(
                        f"""Frame Sequence is not available: {frame_sequence}
                            (specified at index {index} in {frame_selector}""")

        for frame_sequence in actual_frame_sequence_selection:
            sequence = self.get_frame_sequence(frame_sequence)

            available_frames = sorted(sequence.get_complete_frame_sets())

            if frame_selector.frame_selection == '*':
                frame_descriptors.extend(
                    [datatypes.FrameDescriptor(frame_sequence, frame) for frame in available_frames])

                continue
            elif frame_range := frameselection.range_selection(frame_selector.frame_selection, 'frame'):
                try:
                    actual_frame_selection = utils.expand_range(frame_range.as_tuple(), available_frames)
                except utils.MissingItemError as error:
                    raise IOError(
                        f"Frame specified in range as '{error.usage}' is not available: '{error.item}'") from error

                frame_descriptors.extend(
                    [datatypes.FrameDescriptor(frame_sequence, frame) for frame in actual_frame_selection])

                continue
            else:
                actual_frame_selection = [frameselection.normalize_frame_selection_item(frame) for
                                          frame in frameselection.get_items(frame_selector.frame_selection)]

            for index, frame in enumerate(actual_frame_selection):
                if frame not in available_frames:
                    if frame in (incomplete_frame_sets := sequence.get_incomplete_frame_sets()):
                        missing_files = incomplete_frame_sets[frame]

                        raise IOError(
                            f"""Files for frame are missing: {frame_sequence}/{frame}
                                (specified at index {index} in {frame_selector}) Missing: {missing_files}""")
                    else:
                        raise IOError(
                            f"""Frame is not available: {frame_sequence}/{frame}
                                (specified at index {index} in {frame_selector}""")

                frame_descriptors.append(datatypes.FrameDescriptor(frame_sequence, frame))

        return frame_descriptors

    def _get_frame(self, descriptor: datatypes.FrameDescriptor) -> datatypes.Frame:
        return self.get_frame_sequence(descriptor.frame_sequence).get_frame(descriptor.frame)

    def _get_descriptors_from_selections(self, selections: List[str]) -> List[datatypes.FrameDescriptor]:
        selectors = frameselection.get_frame_selectors(selections)

        return self._get_descriptors_from_selectors(selectors)

    def _get_descriptors_from_selectors(
            self,
            selectors: Iterable[frameselection.FrameSelector]) -> List[datatypes.FrameDescriptor]:
        frame_descriptors = []

        for selector in selectors:
            frame_descriptors.extend(self.get_frame_descriptors(selector))

        return frame_descriptors

    def frames(self, frames: Union[List[str], Union[Path, str]], shuffle: bool = False) -> Iterator[datatypes.Frame]:
        """
        Return an iterator for the specified frames.

        To be memory-friendly, the specified frames will be created
        just-in-time. The frames can be specified as either a list of
        selection expressions or as a path to a file with one selection
        expression per line.
        Frame sequences and frames can each be selected by an
        expression containing
            - a single element
            - a list of comma-delimited elements
            - a range of elements similar to slicing works in Python
            - a '*' expressing all available elements.
        Also, 'data_syn' can be used as a frame sequence selection
        element to specify the data_syn frame sequence.
        To get all available frame sequences except for data_syn,
        'data' can be specified as a frame sequence selection
        element.
        A frame sequence selection expression and a frame selection
        expression together form a selection expression. Between both
        frame sequence selection expression and frame selection
        expression, a '/' is placed as the delimiter.

        Examples for valid selection expressions:
            - '1/42'
                Get the 42th frame from the 1st frame sequence
            - '1/[1,2,3]'
                Get the 1st, 2nd and 3th frame from the 1st frame
                sequence
            - '[1,2]/42'
                Get the 42th frame from the 1st and 2nd frame
                sequence each
            - '[1,2]/[4,3]'
                Get the 4th and 3rd frame from the 1st and 2nd frame
                sequence each
            - '[1,2]/*'
                Get all available frames from the 1st and 2nd frame
                sequence each
            - '*/[3,4]'
                Get the 3rd and 4th frame from each the available
                frame sequences
            - '*/*'
                Get all available frames from each the available
                frame sequences
            - 'data_syn/[1,2]'
                Get the 1st and 2nd frame from the data_syn frame
                sequence
            - 42:56:1/1
                Get the 1st frame from each frame sequence between frame
                sequence 42 (inclusive) and 56 (exclusive) by taking steps
                of length 1, i.e. when frame sequence 42 to frame sequence
                55 all are available, 14 frames would be selected
            - 42/::-2
                Get every other frame from frame sequence 42 in reverse
                order from all available frames of frame sequence 42

        If `shuffle` is set to 'False', the frames will be returned
        in exactly the order as specified by the order of selection
        expressions and the order of the elements in those. Otherwise,
        the frames get shuffled. Because random.shuffle() is used to
        shuffle the frames, by setting a seed with random.seed(), the
        frames will be shuffled the same way each time this method is
        called.

        For each specified frame it is made sure that the corresponding
        files exists, even if the frame is only eventually created.
        This can be especially useful when specifying a large number of
        frames and consuming them needs minutes or even hours e.g. in
        a machine-learning context.

        Parameters
        ----------
        frames : Union[List[str], Union[pathlib.Path, str]]
            Either a list of selection expressions or a path to a file
            containing selection expressions.
            The path can be either a string or a Path object.
            If path is not absolute, it is assumed to be relative to
            the root of the dataset.
        shuffle : bool, optional
            If True, the selected frames get shuffled

        Yields
        ------
        Frame
            A frame specified by frames
        """

        if isinstance(frames, list):
            frame_descriptors = self._get_descriptors_from_selections(frames)
        elif isinstance(frames, (Path, str)):
            path = Path(frames)

            if not path.is_absolute():
                path = self._path / path

            frame_selectors = frameselection.load_frame_selectors_from_file(path)
            frame_descriptors = self._get_descriptors_from_selectors(frame_selectors)
        else:
            raise TypeError('frames has to be of type list, pathlib.Path or str')

        if shuffle:
            random.shuffle(frame_descriptors)

        for descriptor in frame_descriptors:
            yield self._get_frame(descriptor)



import pkgutil
import re
from typing import List, Tuple, Callable, Iterable, Dict, TypeVar

from .loader import YcbVideoLoader

T = TypeVar('T')

FRAME_INFO_PATTERN = re.compile(r'^(?P<framesequence>[^:]+): (?P<framecount>[0-9]+)$')


def get_item_ranges(collection: List[T], check_consecution: Callable[[T, T], bool]) -> List[Tuple[int, int]]:
    number_of_items = len(collection)
    if number_of_items == 0:
        return []
    if number_of_items == 1:
        return [(0, 1)]

    ranges = []
    range_start = 0
    previous_item = collection[0]

    for index, item in enumerate(collection[1:], 1):
        if not check_consecution(previous_item, item):
            ranges.append((range_start, index))
            range_start = index

        previous_item = item

    ranges.append((range_start, number_of_items))

    return ranges


def format_in_ranges(items: List[str], check_consecution: Callable[[str, str], bool]) -> [str]:
    formatted_item_ranges = []
    for range_start, range_stop in get_item_ranges(items, check_consecution):
        if range_stop == range_start + 1:
            formatted_item_ranges.append(items[range_start])
        else:
            formatted_item_ranges.append(f"{items[range_start]} - {items[range_stop - 1]}")

    return formatted_item_ranges


def check_sequence_or_frame_consecution(sequence1: str, sequence2: str) -> bool:
    if sequence1.isdigit() and sequence2.isdigit():
        return int(sequence2) == int(sequence1) + 1

    # includes the 'data_syn' sequence
    return False


def read_frame_count_per_frame_sequence() -> Dict[str, int]:
    frames_per_sequence = {}
    frame_info = pkgutil.get_data(__package__, 'frame_info.txt').decode('utf-8').split('\n')

    # skip the newline line at the end of the file
    for entry in frame_info[:-1]:
        # if there is no match, the file would be corrupt
        match = FRAME_INFO_PATTERN.match(entry)
        sequence = match.group('framesequence')
        frame_count = int(match.group('framecount'))

        frames_per_sequence[sequence] = frame_count

    return frames_per_sequence


def print_frame_info(loader: YcbVideoLoader, sequences: Iterable[str], verbosity: int = 0):
    frames_info = loader.frames_info()
    sequence_count = len(frames_info)
    frames_total = read_frame_count_per_frame_sequence()
    missing_sequences = []

    for sequence in sequences:
        if sequence not in frames_info:
            missing_sequences.append(sequence)

    print('Frames')
    print('------')
    print()
    print(f"Frame sequences available: {sequence_count} (total: 92)")
    if verbosity > 0:
        print(f"Available: {format_in_ranges(list(frames_info.keys()), check_sequence_or_frame_consecution)}")
    print(f"Missing: {format_in_ranges(missing_sequences, check_sequence_or_frame_consecution)}")
    print()
    print('Available frame sequences:')
    print('sequence: complete/incomplete/dataset total')
    for sequence, frame_sets in frames_info.items():
        complete_frame_sets = []
        incomplete_frame_sets = []
        for frame_set, missing_files in frame_sets.items():
            if missing_files:
                incomplete_frame_sets.append(frame_set)
            else:
                complete_frame_sets.append(frame_set)
        print(f"{sequence:>8}: {len(complete_frame_sets)}/{len(incomplete_frame_sets)}/{frames_total[sequence]})")
        if incomplete_frame_sets:
            print('Incomplete frame sets:')
            for frame_set in sorted(incomplete_frame_sets):
                print(f"{frame_set}: {frame_sets[frame_set]} missing")
        elif verbosity > 1:
            print('Incomplete frame sets: none')
        if verbosity > 0:
            print('Complete frame sets:')
            print(format_in_ranges(sorted(complete_frame_sets), check_consecution=check_sequence_or_frame_consecution))



from pathlib import Path
from typing import Union, Tuple, Optional, List, Literal


class RangeError(Exception):
    """Base class for exceptions raised in range expansion"""
    pass


class MissingItemError(RangeError):
    """Exception raised when an explicitly named item used as 'start' or 'stop' is missing"""
    def __init__(self, item: str, usage: Literal['start', 'stop'], message: str):
        self.item = item
        self.usage = usage
        self.message = message


def validate_directory_path(path: Union[Path, str]) -> Path:
    path = Path(path)
    if not path.exists():
        raise IOError(f"Path does not exist: {path}")
    if not path.is_dir():
        raise IOError(f"Path is not a directory: {path}")

    return path


def expand_range(item_range: Tuple[Optional[str], Optional[str], Optional[int]], items: List[str]) -> List[str]:
    if not items:
        raise RangeError("No items were given")
    if item_range[2] == 0:
        raise RangeError('Invalid range: Step size must not be 0')

    # step defaults to 1
    # step could be an int != 0 or None at this point
    step = item_range[2] or 1

    start = item_range[0]
    stop = item_range[1]

    if start:
        try:
            start_index = items.index(start)
        except ValueError as error:
            raise MissingItemError(start, 'start', f"Invalid range: Start '{start}' not in {items}") from error
    else:
        start_index = None

    if stop:
        try:
            stop_index = items.index(stop)
        except ValueError as error:
            raise MissingItemError(stop, 'stop', f"Invalid range: Stop '{stop}' not in {items}") from error
    else:
        stop_index = None

    expansion = items[slice(start_index, stop_index, step)]

    # an empty expansion is of no use and would be the result of
    # start > stop with a step size > 0 or
    # stop < start with a step size < 0 or
    # start == stop
    if not expansion:
        raise RangeError(f"Invalid range: Expansion is empty")

    return expansion
